import { BlockVolume, BlockPermutation, EntityTypes, BlockTypes, world, system } from "@minecraft/server";
import { ModalFormData } from "@minecraft/server-ui";

world.beforeEvents.itemUse.subscribe((ev) => {
  const player = ev.source;
  if (ev.itemStack.typeId !== "minecraft:nautilus_shell") return;
  if (ev.source.typeId !== EntityTypes.get("minecraft:player").id) return;
  showCircleForm(player);
})

const typeList = ["keep: 空気の場所のみ設置", "replace: すでにあるブロックも置き換えて設置"];
const HeightLimit = 320;

function showCircleForm(player) {
  const dimension = player.dimension;
  system.run(() => {
    let setForm = new ModalFormData()
      .title("円の設定")
      .textField("中心の座標：\n直接入力", "例）10 0 10   [~]入力ok")
      .toggle("現在地点にする", true)
      .toggle("中心を埋める", false)
      .slider("高さ", 1, 32, 1)
      .slider("直径の大きさ", 8, 64, 1)
      .textField("設置するブロックID", "例）quartz_block ※入力無しはクォーツブロック")
      .dropdown("設置方法", typeList);

    setForm.show(player).then(res => {
      if (res.canceled) return;
      const resInputLocation = res.formValues[0];
      const resCurrentLocation = res.formValues[1];
      const resFillFlg = res.formValues[2];
      const resHeight = res.formValues[3];
      const resDiameter = res.formValues[4];
      const resBlock = res.formValues[5];
      const resSetType = res.formValues[6];

      if (resCurrentLocation && resInputLocation) {
        player.onScreenDisplay.setActionBar(`座標を入力時は§e現在地点をOFFにしてください`);
        return;
      }
      if (resBlock && !BlockTypes.get(resBlock)) {
        player.onScreenDisplay.setActionBar(`§e一致するブロックがありません！`);
        return;
      }

      let centerX = 0;
      let centerY = 0;
      let centerZ = 0;
      let centerXYZ = [ 0, 0, 0 ];
      const { x, y, z } = player.location;
      centerXYZ[0] = Math.floor(Number(x));
      centerXYZ[1] = Math.floor(Number(y));
      centerXYZ[2] = Math.floor(Number(z));

      if (!resCurrentLocation && !resInputLocation) { inputError(player); return; }
      if (resInputLocation) {
        let InputLocations = resInputLocation.split(" ");
        if (InputLocations.length !== 3) { inputError(player); return; }

        for (let i = 0; i < InputLocations.length; i++) {
          if (isNaN(InputLocations[i]) && !InputLocations[i].startsWith("~")) {
            inputError(player);
            return;
          }
          if (InputLocations[i].startsWith("~") 
            && InputLocations[i].length !== 1 
            && isNaN(InputLocations[i].replace("~", ""))) {
              inputError(player);
              return;
            }

          if (!isNaN(InputLocations[i])) {
            centerXYZ[i] = Number(InputLocations[i]);
          } else if (InputLocations[i].startsWith("~") && InputLocations[i].length !== 1 ) {
            centerXYZ[i] = centerXYZ[i] + Number(InputLocations[i].replace("~", ""));
          }
        }
      }
      centerX = centerXYZ[0];
      centerY = centerXYZ[1];
      centerZ = centerXYZ[2];

      if (centerY + resHeight >= HeightLimit) {
        player.onScreenDisplay.setActionBar(`§e設置可能な高さ:${HeightLimit} を超えています。`);
        return;
      }

      let radius = resDiameter / 2;
      const block = resBlock ? resBlock : "quartz_block";
      const options = resSetType == 0 ?
        { matchingBlock: BlockPermutation.resolve("minecraft:air") }
        : {};

      for (let angle = 0; angle < 360; angle++) {
        var cX;
        var cZ;
        if (resDiameter % 2 == 1) {
          cX = centerX + 0.5 + Math.cos(angle * (Math.PI / 180)) * radius;
          cZ = centerZ + 0.5 + Math.sin(angle * (Math.PI / 180)) * radius;

          if (cX > centerX) cX -= (resDiameter + 25.5) ** 2 * 0.00005;
          if (cX < centerX) cX += (resDiameter + 25.5) ** 2 * 0.00005;
          if (cZ > centerZ) cZ -= (resDiameter + 25.5) ** 2 * 0.00005;
          if (cZ < centerZ) cZ += (resDiameter + 25.5) ** 2 * 0.00005;

          const excHandling1 = [25, 33, 35, 37, 39, 59];
          if (cX < centerX && excHandling1.includes(resDiameter)) cX -= 0.025;
          if (cZ < centerZ && excHandling1.includes(resDiameter)) cZ -= 0.025;

          const excHandling2 = [49];
          if (cX < centerX && excHandling2.includes(resDiameter)) cX -= 0.01;
          if (cZ < centerZ && excHandling2.includes(resDiameter)) cZ -= 0.01;

        } else {
          cX = centerX + Math.cos(angle * (Math.PI / 180)) * radius;
          cZ = centerZ + Math.sin(angle * (Math.PI / 180)) * radius;

          if (cX > centerX) cX -= (radius * 0.01);
          if (cX < centerX) cX += (radius * 0.01);
          if (cZ > centerZ) cZ -= (radius * 0.01);
          if (cZ < centerZ) cZ += (radius * 0.01);

          const excHandling1 = [18, 30, 42, 50, 56, 58];
          if (cX < centerX && excHandling1.includes(resDiameter)) cX -= 0.025;
          if (cZ < centerZ && excHandling1.includes(resDiameter)) cZ -= 0.025;
          
          const excHandling2 = [36, 54, 62];
          if (cX < centerX && excHandling2.includes(resDiameter)) cX -= 0.01;
          if (cZ < centerZ && excHandling2.includes(resDiameter)) cZ -= 0.01;
                    
          const excHandling3 = [60];
          if (cX < centerX && excHandling3.includes(resDiameter)) cX -= 0.02;
          if (cZ < centerZ && excHandling3.includes(resDiameter)) cZ -= 0.02;
        }

        let beginLocation = resFillFlg ?
          { x:centerX, y:centerY, z:centerZ }
          : { x:orgRound(cX, 64), y:centerY, z:orgRound(cZ, 64) };
        let endLocation = { x:orgRound(cX, 64), y:centerY + resHeight - 1, z:orgRound(cZ, 64) };
        let fromTo = new BlockVolume(beginLocation, endLocation);

        dimension.fillBlocks(fromTo, block, options);
      }
      if (resFillFlg){
        const tlpLocation = { x:centerX, y:centerY + resHeight, z:centerZ };
        player.teleport(tlpLocation);
      }
      player.sendMessage(`~~~~~~~~~~~~~~~~~~~~~~~~`);
      player.sendMessage(`直径:${resDiameter} の円を作成しました！`);
      player.sendMessage(`中心座標: ${centerX}, ${centerY}, ${centerZ}`);
      player.sendMessage(`~~~~~~~~~~~~~~~~~~~~~~~~`);
    });
  });
}

function inputError(player) {
  player.onScreenDisplay.setActionBar(`§e座標の入力値が正しくありません！`);
}
function orgRound(value, base) {
  return Math.round(value * base) / base;
}